from yta_math_graphic.curve.segment.abstract import CurveSegment
from yta_math_easings import LinearEasing


class LinearCurveSegment(CurveSegment):
    """
    A curve segment that is based on a linear easing
    function to calculate the `y` value.
    """

    def __init__(
        self,
        left_node: 'CurveNode',
        right_node: 'CurveNode'
    ):
        super().__init__(
            left_node = left_node,
            right_node = right_node,
            easing_function = LinearEasing()
        )


from yta_math_graphic.curve.node import CurveNode
from abc import ABC


class CurveSegment(ABC):
    """
    A segment of a curve, including 2 nodes and an
    easing function to define the way we interpolate
    the values in between the 2 nodes.
    """

    @property
    def x_min(
        self
    ) -> float:
        """
        The minimum `x` value in the segment.
        """
        return min(
            (
                node.x
                for node in [self.left_node, self.right_node]
            )
        )
    
    @property
    def x_max(
        self
    ) -> float:
        """
        The maximum `x` value in the segment.
        """
        return max(
            (
                node.x
                for node in [self.left_node, self.right_node]
            )
        )
    
    @property
    def y_min(
        self
    ) -> float:
        """
        The minimum `y` value in the segment.
        """
        return min(
            (
                node.y
                for node in [self.left_node, self.right_node]
            )
        )
    
    @property
    def y_max(
        self
    ) -> float:
        """
        The maximum `y` value in the segment.
        """
        return max(
            (
                node.y
                for node in [self.left_node, self.right_node]
            )
        )
    
    @property
    def distance(
        self
    ) -> float:
        """
        The distance of this segment, considering it as the
        absolute distance in between the two `x` values.
        This value is useful when we need to calculate a
        non-uniform segment distribution.

        The formula:
        ```
        abs(self.x_max - self.x_min)
        ```
        """
        return abs(self.x_max - self.x_min)
    
    def __init__(
        self,
        left_node: 'CurveNode',
        right_node: 'CurveNode',
        easing_function: 'EasingFunction'
    ):
        # TODO: Validate (?)
        # TODO: Maybe 'left' and 'right' is not the best
        # because if we accept unordered, when plotting
        # the 'left' it could be drawn on the right...
        self.left_node: 'CurveNode' = left_node
        """
        The node that is in the left side of the segment.
        """
        self.right_node: 'CurveNode' = right_node
        """
        The node that is in the right side of the segment.
        """
        self._easing_function: 'EasingFunction' = easing_function
        """
        *For internal use only*

        The easing function that will be applined when
        calculating the value in between the 2 nodes.
        """

    def evaluate(
        self,
        progress_local: float
    ) -> float:
        """
        Get the value associated to the local progress
        value provided as `progress_local`, applying
        the easing function defined for this instance.

        The `progress_local` must be a value in the
        `[0.0, 1.0]` range.

        The formula:
        - `self.left_node.y + t * (self.right_node.y - self.left_node.y)`
        """
        # We transform the `t` based on the easing function
        t = self._easing_function(progress_local)

        return self.left_node.y + t * (self.right_node.y - self.left_node.y)

    def get_weight(
        self,
        total_distance: float
    ) -> float:
        """
        Get the weight of this segment based on the 
        `total_distance` of the curve and the distance of
        this segment.

        This method is useful when calculating a 
        proportional distribution in which each segment's
        is different according to its own distance.

        The formula:
        - `self.distance / total_distance`
        """
        return self.distance / total_distance
    
    def is_x_contained(
        self,
        x: float
    ) -> bool:
        """
        Check if the `x` value provided is in between the
        2 nodes of this segment.

        The formula:
        - `self.x_min <= x < self.x_max`
        """
        return self.x_min <= x < self.x_max
    
    def get_progress_local(
        self,
        x: float
    ) -> float:
        """
        Get the local progress of the `x` value provided
        based on the internal distance of the segment.

        The `x` value must be contained in between the
        2 nodes, excluding the one on the right.
        
        The formula:
        - `(x - self.x_min) / self.distance`
        """
        if not self.is_x_contained(x):
            raise Exception('The "x" value provided does not belong to this segment.')
        
        return (x - self.x_min) / self.distance



from yta_math_graphic.curve.node import CurveNode
from typing import Union


"""
TODO: As we always want the nodes in order
for this general Curve concept, we could
force them to be ordered before storing them
when the `add_node` is called, so the `x_min`
and those properties are easier.
"""
class CurveDefinition:
    """
    The definition of a curve.
    """

    @property
    def nodes(
        self
    ) -> list[CurveNode]:
        """
        The list of nodes but ordered by the `x` value.

        This value is cached.
        """
        if self._nodes_ordered is None:
            self._nodes_ordered = tuple(sorted(self._nodes, key = lambda node: node.x))

        return self._nodes_ordered
    
    @property
    def x_min(
        self
    ) -> Union[float, None]:
        """
        The minimum `x` value, as a normalized value in
        the `[0.0, 1.0]` range. Useful value to use for
        the axis and because it is where the progress
        starts.
        """
        return min(
            (
                node.x
                for node in self.nodes
            ),
            default = None
        )
    
    @property
    def x_max(
        self
    ) -> Union[float, None]:
        """
        The maximum `x` value, as a normalized value in
        the `[0.0, 1.0]` range. Useful value to use for
        the axis and because it is where the progress
        ends.
        """
        return max(
            (
                node.x
                for node in self.nodes
            ),
            default = None
        )
    
    @property
    def y_min(
        self
    ) -> Union[float, None]:
        """
        The minimum `y` value, as a normalized value in
        the `[0.0, 1.0]` range.
        """
        return min(
            (
                node.y
                for node in self.nodes
            ),
            default = None
        )
    
    @property
    def y_max(
        self
    ) -> Union[float, None]:
        """
        The maximum `y` value, as a normalized value in
        the `[0.0, 1.0]` range.
        """
        return max(
            (
                node.y
                for node in self.nodes
            ),
            default = None
        )
    
    @property
    def distance(
        self
    ) -> Union[float, None]:
        """
        The total length (as a normalized value) of the
        segments we have in this definition.

        The formula:
        ```
        self.x_max - self.x_min
        ```
        """
        return (
            self.x_max - self.x_min
            if self.x_max is not None else
            None
        )
    
    @property
    def number_of_segments(
        self
    ) -> int:
        """
        The number of segments that the curve has according
        to the amount of nodes included.
        """
        return max(0, len(self._nodes) - 1)
    
    def __init__(
        self,
        x_value_range: tuple[float, float] = (0.0, 1.0),
        y_value_range: tuple[float, float] = (0.0, 1.0)
    ):
        self._x_value_range = x_value_range
        """
        The range of values that will be applied to 
        transform the `x` value obtained (which is
        internally normalized) into the real expected
        value.
        """
        self._y_value_range = y_value_range
        """
        The range of values that will be applied to 
        transform the `y` value obtained (which is
        internally normalized) into the real expected
        value.
        """
        self._nodes: list[CurveNode] = []
        """
        *For internal use only*

        The nodes in the curve as they were added to it.
        """
        self._nodes_ordered: Union[list[CurveNode], None] = None
        """
        *For internal use only*

        The list of nodes but ordered by the `x` axis from
        left to right.
        """

    def get_node_at(
        self,
        x: float
    ) -> Union[CurveNode, None]:
        """
        Get the node, if existing, at the `x` position
        provided.
        """
        return next(
            (
                node
                for node in self.nodes
                if node.x == x
            ),
            None
        )

    def add_node(
        self,
        x: float,
        y: float
    ) -> 'CurveDefinition':
        """
        Add a new node to the `x` provided with the `y`
        associated value if there is no other node 
        including that `x` value.

        This will clean the cache.
        """
        if self.get_node_at(x) is not None:
            raise Exception('There is a node in the `x` position provided.')
        
        self._nodes.append(CurveNode(x, y))
        self._nodes_ordered = None

        return self


from yta_math_graphic.curve.segment.distribution.abstract import WeightedSegmentDistribution
from yta_math_graphic.curve.segment.abstract import CurveSegment


class UniformWeightedSegmentDistribution(WeightedSegmentDistribution):
    """
    A weighted segment distribution in which each
    segment's weight is the same, ignoring the specific
    distance in each segment.
    """

    @staticmethod
    def get_weights(
        segments: list['CurveSegment']
    ) -> list[float]:
        number_of_segments = len(segments)

        return [1.0 / number_of_segments] * number_of_segments
        


from yta_math_graphic.curve.segment.distribution.abstract import WeightedSegmentDistribution
from yta_math_graphic.curve.segment.abstract import CurveSegment


class ProportionalSegmentDistribution(WeightedSegmentDistribution):
    """
    A weighted segment distribution in which seach segment
    weights are proportional to their real distance, which
    means that the longer segments will have more weight
    when calculating the distribution.
    """

    @staticmethod
    def get_weights(
        segments: list['CurveSegment']
    ) -> list[float]:
        total_distance = sum(segment.distance for segment in segments)

        return [
            segment.get_weight(total_distance)
            for segment in (segments)
        ]



from yta_math_graphic.curve.segment.linear import LinearCurveSegment
from yta_math_graphic.curve.segment.smootherer import SmoothererCurveSegment


__all__ = [
    'LinearCurveSegment',
    'SmoothererCurveSegment'
]


from yta_validation.parameter import ParameterValidator
from dataclasses import dataclass


@dataclass
class CurveNode:
    """
    A node within a curve.
    """

    def __init__(
        self,
        x: float,
        y: float
    ):
        ParameterValidator.validate_mandatory_float('x', x)
        ParameterValidator.validate_mandatory_float('y', y)

        self.x: float = x
        """
        The `x` position of the node.
        """
        self.y: float = y
        """
        The `y` position of the node.
        """


from yta_math_graphic.curve.segment.distribution.abstract import WeightedSegmentDistribution
from yta_math_graphic.curve.segment.distribution.uniform import UniformWeightedSegmentDistribution
from yta_math_graphic.curve.segment.factory import SegmentFactory
from yta_math_graphic.curve.evaluator import CurveEvaluator
from yta_math_graphic.curve.definition import CurveDefinition
from yta_programming.decorators.requires_dependency import requires_dependency
from yta_validation.parameter import ParameterValidator
from typing import Union


class Curve:
    """
    A curve, representing both `x` and `y` axis in a
    normalized range (`[0.0, 1.0]`).

    There are no mandatory nodes for this curve, which
    means that it could have not any node at the `x=0.0`
    nor `x=1.0`. Check `ParameterCurve` if you want them
    to be mandatory.
    """
    
    @property
    def nodes(
        self
    ) -> list['CurveNode']:
        """
        The list of nodes ordered by the `x` value.
        """
        return self._definition.nodes
    
    @property
    def segments(
        self
    ) -> list['CurveSegment']:
        """
        The list of segments ordered by the `x` value.
        """
        return self._evaluator._segments
    
    @property
    def x_min(
        self
    ) -> Union[float, None]:
        """
        The minimum `x` value. Useful value to use for
        the axis and because it is where the progress
        starts.
        """
        return self._definition.x_min
    
    @property
    def x_max(
        self
    ) -> Union[float, None]:
        """
        The maximum `x` value. Useful value to use for
        the axis and because it is where the progress
        ends.
        """
        return self._definition.x_max
    
    @property
    def y_min(
        self
    ) -> Union[float, None]:
        """
        The minimum `y` value.
        """
        return self._definition.y_min
    
    @property
    def y_max(
        self
    ) -> Union[float, None]:
        """
        The maximum `y` value.
        """
        return self._definition.y_max
    
    @property
    def distance(
        self
    ) -> Union[float, None]:
        """
        The distance (normalized) of the whole curve.
        """
        return self._definition.distance
    
    @property
    def x_value_range(
        self
    ) -> tuple[float, float]:
        """
        The range of values that will be applied to
        transform the `x` value obtained (which is
        internally normalized) into the real expected
        value.
        """
        return self._definition._x_value_range

    @property
    def y_value_range(
        self
    ) -> tuple[float, float]:
        """
        The range of values that will be applied to
        transform the `y` value obtained (which is
        internally normalized) into the real expected
        value.
        """
        return self._definition._y_value_range

    def __init__(
        self,
        x_value_range: tuple[float, float] = (0.0, 1.0),
        y_value_range: tuple[float, float] = (0.0, 1.0),
        distribution: WeightedSegmentDistribution = UniformWeightedSegmentDistribution(),
        segment_factory: SegmentFactory = SegmentFactory(),
    ):
        """
        The value range is the range of values in which 
        each axis will be forced to belong to and all
        the values will be in.

        The curve works internally with normalized values
        (in the `[0.0, 1.0]` range), but they are
        denormalized to the real values that fit this
        range to be able to use it in a real environment.

        If you want to represent a parameter, you can use
        `y_value_range=(0, 255)` and all the values will be
        inside that range, being transformed like this:
        - `0.0 -> 0`
        - `1.0 -> 255`
        """
        ParameterValidator.validate_mandatory_subclass_of('distribution', distribution, WeightedSegmentDistribution)
        ParameterValidator.validate_mandatory_instance_of('segment_factory', segment_factory, SegmentFactory)
        # TODO: Validate 'x_value_range'
        # TODO: Validate 'y_value_range'
        
        self._definition = CurveDefinition(
            x_value_range = x_value_range,
            y_value_range = y_value_range
        )
        """
        *For internal use only*

        The definition of the curve.
        """
        self._distribution = distribution
        """
        *For internal use only*

        The distribution of the curve.
        """
        self._segment_factory = segment_factory
        """
        *For internal use only*

        The segment factory of the curve.
        """
        self._evaluator: CurveEvaluator = None
        """
        *For internal use only*

        The evaluator of the curve.
        """

        self._reset_cache()

    def add_node(
        self,
        x: float,
        y: float
    ) -> 'Curve':
        """
        Add a new node to the real `x` provided with the
        also real `y` value associated to it. 
        
        The `x` and `y` values provided must be inside
        the value ranges provided for this instance.
        """
        ParameterValidator.validate_mandatory_number_between('x', x, self.x_value_range[0], self.x_value_range[1])
        ParameterValidator.validate_mandatory_number_between('y', y, self.y_value_range[0], self.y_value_range[1])

        # Normalize the 'y' before storing it
        x = self._evaluator._normalize_x(x)
        y = self._evaluator._normalize_y(y)

        self._definition.add_node(x, y)
        # Invalidate cache
        self._evaluator = None
        self._reset_cache()

        return self

    def _reset_cache(
        self
    ) -> 'Curve':
        """
        *For internal use only*

        Reset the cache by reseting the evaluator.
        """
        self._evaluator = CurveEvaluator(
            self._definition,
            self._distribution,
            self._segment_factory
        )

        return self

    def evaluate(
        self,
        progress_normalized: float
    ) -> float:
        """
        Get the real `y` value associated to the global
        `progress_normalized` provided.

        The value associated will be inside the value
        range provided for this curve as it is a real
        value.
        """
        if self._evaluator is None:
            self._reset_cache()

        return self._evaluator.evaluate(
            progress_normalized = progress_normalized
        )
    
    def evaluate_at(
        self,
        x: float
    ) -> float:
        """
        Get the real `y` value associated to the real `x`
        value provided.

        The value associated will be inside the value
        range provided for this curve.
        """
        if self._evaluator is None:
            self._reset_cache()

        return self._evaluator.evaluate_at(
            x = x
        )

    @requires_dependency('matplotlib', 'yta_math', 'matplotlib')
    def plot(
        self,
        number_of_samples_per_segment: int = 100,
        do_draw_nodes: bool = True,
        do_draw_segments: bool = True
    ) -> None:
        """
        *Requires optional dependency `matplotlib`*

        Displays the curve in a normalized space
        `[0.0, 1.0] -> [0.0, 1.0]`.
        """
        import matplotlib.pyplot as plt
        import numpy as np

        plt.xlim(self.x_value_range[0], self.x_value_range[1])
        plt.ylim(self.y_value_range[0], self.y_value_range[1])
        plt.axhline(0.0, linewidth = 1)
        plt.axvline(0.0, linewidth = 1)
        plt.grid(True)

        nodes = self.nodes

        if do_draw_segments:
            for node_left, node_right in zip(nodes[:-1], nodes[1:]):
                xs = np.linspace(node_left.x, node_right.x, number_of_samples_per_segment)
                ys = [
                    self.evaluate_at(self._evaluator._denormalize_x(x))
                    for x in xs
                ]
                # Denormalize to print
                xs = [
                    self._evaluator._denormalize_x(x)
                    for x in xs
                ]
                plt.plot(xs, ys, linewidth = 2)

        if do_draw_nodes:
            x_nodes = [
                self._evaluator._denormalize_x(node.x)
                for node in nodes
            ]
            y_nodes = [
                self._evaluator._denormalize_y(node.y)
                for node in nodes
            ]

            plt.scatter(
                x_nodes,
                y_nodes,
                s = 80,
                facecolors = 'white',
                edgecolors = 'black',
                zorder = 3
            )

        plt.xlabel('Progress (normalized)')
        plt.ylabel('Value (normalized)')
        plt.title('Curve')

        plt.show()


# TODO: Maybe move to the editor instead (?)
class ParameterCurve(Curve):
    """
    A curve that always includes values for the initial
    (`x_normalized=0.0`) and final (`x_normalized=1.0`)
    nodes, forcing it to have real values for all the
    values in the internal and normalized range
    `[0.0, 1.0]`.

    It will be initialized with 2 mandatory nodes, at
    `x=0.0` and at `x=1.0` having the values provided
    in the `__init__` method.

    This curve is perfect to be applied in edition to
    set all the parameter values we want.
    """

    def __init__(
        self,
        y_start: float,
        y_end: float,
        x_value_range: tuple[float, float] = (0.0, 1.0),
        y_value_range: tuple[float, float] = (0.0, 1.0),
        distribution: WeightedSegmentDistribution = UniformWeightedSegmentDistribution(),
        segment_factory: SegmentFactory = SegmentFactory(),
    ):
        """
        The `y_start` and `y_end` values must be according
        to the `y_value_range` provided.
        """
        super().__init__(
            x_value_range = x_value_range,
            y_value_range = y_value_range,
            distribution = distribution,
            segment_factory = segment_factory
        )

        # We force to add the first and last node
        self.add_node(
            x = x_value_range[0],
            y = y_start
        ).add_node(
            x = x_value_range[1],
            y = y_end
        )


from yta_math_graphic.curve.node import CurveNode
from yta_math_graphic.curve.segment.abstract import CurveSegment
from yta_math_graphic.curve.segment import LinearCurveSegment, SmoothererCurveSegment


class SegmentFactory:
    """
    A factory to create curve segments.
    """

    def create(
        self,
        left_node: CurveNode,
        right_node: CurveNode
    ) -> CurveSegment:
        """
        Create a new `CurveSegment` including the `left_node`
        and the `right_node` provided.
        """
        # TODO: Why is it only Linear here and how to
        # make it customizable (?)
        return SmoothererCurveSegment(
            left_node = left_node,
            right_node = right_node
        )
        # return LinearCurveSegment(
        #     left_node = left_node,
        #     right_node = right_node
        # )



from yta_math_graphic.curve.segment.abstract import CurveSegment
from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod
from typing import Union


class WeightedSegmentDistribution(ABC):
    """
    A segment distribution in which the different
    segments have different weights.

    The distribution of the curve, that defines how
    a global `progress_normalized` value is turned into this:
    - `segment_index`
    - `progress_local`
    """

    @abstractmethod
    def get_weights(
        self,
        segments: list['CurveSegment']
    ) -> list[float]:
        """
        Get the weights for the different segments according
        to the curve segments provided.

        This method must be implemented by each specific
        class.
        """
        pass

    def get_y_from_progress(
        self,
        progress_normalized: float,
        segments: list[CurveSegment]
    ) -> float:
        """
        Transform the global `progress_normalized`
        provided into the `y` normalized value associated
        to it, looking for the segments (inside of the
        `segments` list provided) the progress belongs
        to and calculating it.
        """
        segment, progress_local_normalized = self._get_segment_and_local_progress_from_progress(
            progress_normalized = progress_normalized,
            segments = segments
        )

        return segment.evaluate(progress_local_normalized)

    def get_y_from_x(
        self,
        x: float,
        segments: list[CurveSegment]
    ) -> float:
        """
        Transform the `x` normalized value into the `y` 
        normalized value associated to that value, looking
        for the exact segment that contains it (into the
        `segments` provided) and evaluating the local
        progress that is calculated according to the `x`
        given.
        """
        segment, progress_local_normalized = self._get_segment_and_local_progress_from_x(
            x = x,
            segments = segments
        )

        return segment.evaluate(progress_local_normalized)

    def _get_segment_and_local_progress_from_x(
        self,
        x: float,
        segments: list[CurveSegment]
    ) -> tuple[CurveSegment, float]:
        """
        Transform the `x` normalized value provided
        into the local and normalized progress and 
        the corresponding segment, returninga tuple
        containing:
        - `segment`
        - `progress_local_normalized`
        """
        segment = self._get_segment_from_x(
            x = x,
            segments = segments
        )

        if segment is None:
            raise Exception(f'No segment for the x={str(x)} provided.')

        x0 = segment.x_min
        x1 = segment.x_max

        progress_local_normalized = (x - x0) / (x1 - x0)

        return (segment, progress_local_normalized)
    
    def _get_segment_and_local_progress_from_progress(
        self,
        progress_normalized: float,
        segments: list[CurveSegment]
    ) -> tuple[CurveSegment, float]:
        """
        Get the segment the provided global `progress_normalized`
        belongs to, and also de local progress according to that
        segment.

        This method will return a tuple containing:
        - `segment`
        - `progress_local`
        """
        ParameterValidator.validate_number_between('progress_normalized', progress_normalized, 0.0, 1.0)
        ParameterValidator.validate_mandatory_list_of_these_instances('segments', segments, CurveSegment)

        segment_weights = self.get_weights(segments)

        total_weight = sum(segment_weights)
        if total_weight <= 0.0:
            raise Exception('Total segment weight must be > 0')

        target = progress_normalized * total_weight
        accumulated_weight = 0.0

        for index, weight in enumerate(segment_weights):
            next_accumulated_weight = accumulated_weight + weight

            if (
                target <= next_accumulated_weight or
                index == len(segment_weights) - 1
            ):
                # We found the segment corresponding to the global
                # `progress_normalized` provided
                local = target - accumulated_weight
                progress_local = (
                    local / weight
                    if weight > 0 else
                    0.0
                )

                return segments[index], progress_local

            accumulated_weight = next_accumulated_weight

        return (segments[-1], 1.0)

    def _get_segment_from_x(
        self,
        x: float,
        segments: list[CurveSegment]
    ) -> Union[CurveSegment, None]:
        """
        *For internal use only*

        Get the segment in which the `x` value provided fits,
        that is by applying the `[x_min, x_max)` range for
        all the segments and `[x_min, x_max]` for the last
        one.
         
        For the last segment:
        - `if segment.x_min <= x <= segment.x_max`

        For the rest:
        - `if segment.x_min <= x < segment.x_max`
        """
        for i, segment in enumerate(segments):
            if i == len(segments) - 1:
                if segment.x_min <= x <= segment.x_max:
                    return segment
            else:
                if segment.x_min <= x < segment.x_max:
                    return segment
                
        return None


from yta_math_graphic.curve.definition import CurveDefinition
from yta_math_graphic.curve.segment.distribution.abstract import WeightedSegmentDistribution
from yta_math_graphic.curve.segment.factory import SegmentFactory


class CurveEvaluator:
    """
    The evaluator of a curve, which is able to
    calculate the values according to the progress
    or `x` value provided.
    """

    @property
    def _nodes(
        self
    ) -> list['CurveNode']:
        """
        The nodes of the curve, ordered by the `x` value.
        """
        return self._definition.nodes

    def __init__(
        self,
        definition: CurveDefinition,
        distribution: WeightedSegmentDistribution,
        segment_factory: SegmentFactory
    ):
        self._definition = definition
        """
        *For internal use only*

        The definition of the curve.
        """
        self._segments = [
            segment_factory.create(self._nodes[i], self._nodes[i + 1])
            for i in range(len(self._nodes) - 1)
        ]
        """
        *For internal use only*

        The segments of the curve, created based on the
        `segment_factory` provided, and including the
        nodes ordered by the `x` value.
        """
        # TODO: Maybe rename to '_weighted_distribution'
        self._distribution = distribution
        """
        *For internal use only*

        The distribution of the segments to apply in the
        curve.
        """

    def evaluate(
        self,
        progress_normalized: float
    ) -> float:
        """
        Get the value associated to the global
        `progress_normalized` provided, that will be
        transformed into the `y` value associated to it.

        The value returned will be the real value according
        to the `value_range` set in the `Curve` instance.
        """
        y_normalized = self._distribution.get_y_from_progress(
            progress_normalized = progress_normalized,
            segments = self._segments
        )

        # We force it to be denormalized
        do_denormalize: bool = True

        return (
            self._denormalize_y(y_normalized)
            if do_denormalize else
            y_normalized
        )
    
    def evaluate_at(
        self,
        x: float
    ) -> float:
        """
        Get the value associated to the `x` real value
        provided, that will be transformed into the `y`
        value associated to it.

        The value returned will be the real value according
        to the `value_range` set in the `Curve` instance.
        """
        x_normalized = self._normalize_x(x)

        y_normalized = self._distribution.get_y_from_x(
            x = x_normalized,
            segments = self._segments
        )

        # We force it to be denormalized
        do_denormalize: bool = True

        return (
            self._denormalize_y(y_normalized)
            if do_denormalize else
            y_normalized
        )
    
    # Not used by now
    def _normalize_x(
        self,
        x: float
    ) -> float:
        """
        *For internal use only*

        Transform the not normalized `x` value into the
        normalized value according to the value range
        defined for the curve.
        """
        return _normalize(
            value = x,
            value_range = self._definition._x_value_range
        )
    
    def _denormalize_x(
        self,
        x_normalized: float
    ) -> float:
        """
        *For internal use only*

        Transform the `x_normalized` value into the real
        value according to the value range defined for the
        curve.
        """
        return _denormalize(
            value = x_normalized,
            value_range = self._definition._x_value_range
        )

    # Not used by now
    def _normalize_y(
        self,
        y: float
    ) -> float:
        """
        *For internal use only*

        Transform the not normalized `y` into the
        normalized value according to the value range
        defined for the curve.
        """
        return _normalize(
            value = y,
            value_range = self._definition._y_value_range
        )
    
    def _denormalize_y(
        self,
        y_normalized: float
    ) -> float:
        """
        *For internal use only*

        Transform the `y_normalized` value into the real
        value according to the value range defined for the
        curve.
        """
        return _denormalize(
            value = y_normalized,
            value_range = self._definition._y_value_range
        )
    
def _normalize(
    value: float,
    value_range: tuple[float, float]
) -> float:
    """
    *For internal use only*

    Transform the not normalized `value` into the
    normalized value according to the `value_range`
    provided.
    """
    min_value, max_value = value_range

    return (value - min_value) / (max_value - min_value)
    
def _denormalize(
    value: float,
    value_range: tuple[float, float]
) -> float:
    """
    *For internal use only*

    Transform the normalized `value` into the real
    value according to the `value_range` provided.
    """
    min_value, max_value = value_range

    return min_value + value * (max_value - min_value)
        




from yta_math_graphic.curve.segment.distribution.uniform import UniformWeightedSegmentDistribution
from yta_math_graphic.curve.segment.distribution.proportional import ProportionalSegmentDistribution


"""
TODO: Continue implementing new options:
1. Ease / EasingSegment
    EaseIn
    EaseOut
    EaseInOut
    SmoothStep
    S-Curve
2. BezierCurveSegment
3. Hermite / CubicSegment
4. Step / HoldSegment
5. Extrapolated / OvershootSegment
"""

__all__ = [
    'UniformWeightedSegmentDistribution',
    'ProportionalSegmentDistribution'
]